#[allow(unused_must_use)]
pub mod shell {
    use regex::Regex;
    use std::io::Stdout;
    use std::process::Command;
    use std::io::stdin;
    use std::io::{stdout};
    use std::io::Write;
    use std::path::Path;
    use std::env;
    use std::str;
    use std::fs::File;
    use std::process::Stdio;
    use termion::event::Key;
    use termion::input::TermRead;
    use termion::raw::IntoRawMode;
    use termion::raw::RawTerminal;
    use termion::{color, cursor};
    use termion;
    use unicode_segmentation::UnicodeSegmentation;
    use std::collections::HashMap;

    mod history;
    use crate::shell::history::*;

    mod autocomplete;
    use crate::shell::autocomplete::*;

    mod config;
    use crate::shell::config::*;

    //used for home directory
    extern crate dirs;
    extern crate shell_words;
    extern crate ctrlc;

    ///Stores essential terminal variables
    struct TermElements {
        //Stores currently typed command
        command: String,
        //Current cursor position in the line
        cur_pos: usize,
        //Max position of cursor in the line
        max_pos: usize,
        stdout: RawTerminal<Stdout>,
        //Current history number
        current_number: i32,
        //Configuration
        conf: SqishConf,
    }

    ///Used to unify command output and error handling 
    ///across functions
    struct CmdOutput {
        outp: String,
        rc: i16,
    }

    impl CmdOutput {
        fn from_values(output: String, rc: i16) -> CmdOutput {
            let myoutp = CmdOutput {
                outp: output,
                rc: rc,
            };
            return myoutp;
        }
        fn new_empty() -> CmdOutput {
            let outp = CmdOutput {
                outp: String::new(),
                rc: 0,
            };
            return outp;
        }
    }

    fn export_var(command: String) -> Result<CmdOutput, CmdOutput> {
        let cmd_split = command.split('=').collect::<Vec<&str>>();
        if cmd_split.len() < 2 {
            return Err(CmdOutput::from_values(
                    String::from("Please a format such as : VAR=value"),
                    255));
        }
        let key = String::from(cmd_split[0].to_uppercase());
        let val = String::from(cmd_split[1]);
        let mut map = HashMap::new();
        map.insert(key, val);
        set_envvars(&map);
        return Ok(CmdOutput::new_empty());
    }

    fn change_dir(args: &Vec<String>, previous_command: &mut Option<Stdio>) -> Result<CmdOutput, CmdOutput> {
        let homedir = dirs::home_dir().unwrap().into_os_string().into_string().unwrap();
        //let new_dir = args.peekable().peek().map_or(homedir.as_str(), |x| *x);
        let mut new_dir = String::new();
        if args.len() > 0 {
            *&mut new_dir = args.join(" ");
        } else {
            *&mut new_dir = String::from(homedir.as_str());
        }

        let root = Path::new(&new_dir);
        if let Err(e) = env::set_current_dir(&root) {
            return Err(CmdOutput::from_values(
                    format!("Error setting directory to {}, got {:?}",
                                 new_dir, e), -1));
        };
        *previous_command = None;
        return Ok(CmdOutput::new_empty());
    }

    fn print_hist(previous_command: &mut Option<Stdio>, command_next: bool) -> Result<CmdOutput, CmdOutput> {
        //let stdout = Stdio::inherit();
        let res = get_history();
        match res {
            Ok(r) => {
                let filepath = dirs::home_dir().unwrap().join(".history.sqsh");
                let file = File::open(filepath).unwrap();
                *previous_command = Some(Stdio::from(file));
                //if !commands.peek().is_some() {
                if !command_next {
                    print!("{}", r);
                }
            },
            Err(e) => {
                return Err(CmdOutput::from_values(format!("Could not print history, got {:?}", e), 255));
            },
        }
        return Ok(CmdOutput::new_empty());
    }

    fn read_hotkeys(hotkeys: &HashMap<String, String>) -> HashMap<String, String> {
        //Read the env vars
        //and add hotkeys as we find them
        let mut hm = hotkeys.clone();
        for (key, value) in env::vars() {
            if key.starts_with("SQISH_HK_") && key.chars().count() == 10 {
                let letter = key.chars().last().unwrap();
                &hm.insert(String::from(letter).to_lowercase(), value.clone());
            }
        }
        return hm;
    }

    fn set_hotkey(args: Vec<String>) -> Result<String, String> {
        if args.len() < 2 { return Err(String::from("Not enough arguments given for hotkey creation.")); }
        let letter = args
            .first()
            .unwrap()
            .to_lowercase();
        let letter = letter
            .chars()
            .nth(0)
            .ok_or('a')
            .unwrap();
        if !letter.is_alphabetic() { return Err(format!("Sign given is not alphabetic")); }
        let mut command = args[1..]
            .join(" ");
            
        command.push_str("\n");

        let mut hm = HashMap::new();
        let key = format!("SQISH_HK_{}", letter.to_uppercase());
        hm.insert(key, command); 
        set_envvars(&hm);

        return Ok(format!("Hotkey set for Alt+{letter}"));
    }

        
    fn process_line(input: &str) -> Result<CmdOutput, CmdOutput> {
        let mut commands = input.trim().split("|").peekable();
        let mut previous_command = None;
        let mut resultat = String::new();
        let mut rc: i16 = 0;

        while let Some(command) = commands.next()  {
            let parts = match shell_words::split(&command.trim()) {
                Ok(w) => w,
                Err(e) => { 
                    return Err(CmdOutput::from_values(format!("Could not parse command {:?}", e), -1));
                },
            };
            if parts.len() < 1 {
                return Err(CmdOutput::from_values(String::from("Could not parse command"), -1));
            }
            let command = parts[0].as_str();
            let args = Vec::from(&parts[1..]);

            match command {
                "cd" => {
                    change_dir(&args, &mut previous_command)?;
                    print!("\r\n");
                },
                "history" => {
                    let next = commands.peek().is_some();
                    print_hist(&mut previous_command, next)?;
                },
                "export" => {
                    export_var(args.join(" "))?;
                    print!("\r\n");
                },
                "hk" => {
                    match set_hotkey(args) {
                        Ok(_) => { print!("\r\nHotkey registered.\r\n"); },
                        Err(e) => { 
                            let line = format!("\r\n{e}\r\n");
                            print!("{}", &line);
                        },
                    };
                }
                command => {
                    if commands.peek().is_some() {
                        let stdin  = match previous_command {
                            None => Stdio::inherit(),
                            Some(o) => o,
                        };
                        let stdout = Stdio::piped();
                    
                        let output = Command::new(command)
                            .args(args)
                            .stdout(stdout)
                            .stdin(stdin)
                            .spawn();

                        match output {
                            Ok(o) => previous_command = Some(Stdio::from(o.stdout.unwrap())),
                            Err(e) => {
                                previous_command = None;
                                eprintln!(" Err: {}", e);
                            },
                        };
                    } else {
                        let stdin = match previous_command {
                            None => Stdio::inherit(),
                            Some(o) => o,
                        };

                        print!("\r\n");
                      
                        let child = match Command::new(command)
                            .args(args)
                            .stdin(stdin)
                            .spawn() {
                                Ok(h) => h,
                                Err(_) => {
                                    let err_string = format!("Not found : {}", command);
                                    return Err(CmdOutput::from_values(err_string, -1));
                                },
                            };

                        let output = child
                            .wait_with_output()
                            .expect("Failed to wait on child");
                        let status = output.status
                            .code()
                            .unwrap_or(255)
                            .to_string();
                        previous_command = None;
                        let format_res = format!("{}", 
                                                 status);
                        *&mut rc = format_res.parse::<i16>().unwrap();
                        *&mut resultat = String::from_utf8(output.stdout).unwrap();
                        //let _ = &mut resultat.push_str(&format_res);
                    }
                },
            };
        }
        //return resultat;
        return Ok(CmdOutput::from_values(resultat, rc));
    }

    fn replace_signs(line: &String) -> String {
        let mut ayo = String::from(line);
        if ayo.contains('~') {
            let homedir = dirs::home_dir().unwrap().into_os_string().into_string().unwrap();
            let result = str::replace(&ayo, "~", &homedir);
            ayo = String::from(result);
        }
        
        if ayo.contains("!!") {
            let prev_comm = get_hist_from_number(&get_history_number().unwrap()).unwrap();
            let result = str::replace(&ayo, "!!", &prev_comm);
            ayo = String::from(result);
        }

        return ayo;
    }

    fn handle_input(line: &str) -> Result<CmdOutput, CmdOutput> {
        //history callback
        if (line.len() > 0) && (line.starts_with('!'))  {
            let command_found = treat_history_callback(line);
            match command_found {
                Some(c) => {
                    write_to_history(&c);
                    return process_line(&c);
                    },
                None => return Err(CmdOutput::from_values(
                        String::from("No such value"),
                        255)),
            };
        //Regular command
        } else if line.len() > 0 {
            write_to_history(line);
            return process_line(line);
        //Nothing
        } else {
            //Command was empty, new line
            return Ok(CmdOutput::new_empty());
        }
    }

    fn delete_current_cmd(current_cmd: String) -> String {
        //Deletes currently written command
       
        let mut stdout = stdout().into_raw_mode().unwrap();
        //write!(stdout, "{}", cursor::Save);
        for _c in current_cmd.graphemes(true) {
            write!(stdout, "\x1b[D").unwrap();
            write!(stdout, "\x1b[K").unwrap();
        }
        stdout.flush();
        //write!(stdout, "{}", cursor::Restore);
        return String::new();
    }

    fn get_cmd_curs_pos(command: &String) -> (usize, usize) {
        let mut max_pos: usize = 0; 
        let mut current_pos: usize = 0; 

        for _c in command.graphemes(true) {
            max_pos += 1;
            current_pos += 1;
        }

        return (max_pos, current_pos);
    }

    fn clear_line(max_pos: &usize, current_pos: &usize) {
        //clears currently written command from the screen
        //...NOT from the variable !
        let mut stdout = stdout().into_raw_mode().unwrap();
        let chars_left = max_pos - current_pos;
        if current_pos < max_pos {
            write!(stdout, "{}", cursor::Right(chars_left as u16));
        }
        for _i in 0..*max_pos {
            write!(stdout, "{}", cursor::Left(1));
            write!(stdout, "\x1b[K").unwrap();
        }
        //write!(stdout, "{}", cursor::Left(1));
        //write!(stdout, "\x1b[K").unwrap();
        stdout.flush();
    }

    fn transform_alias(conf: &SqishConf, full_command: String) -> String {
        let binding = full_command.clone();
        let mut pieces = binding.split_whitespace()
                                    .collect::<Vec<&str>>();
        let command_temp = match pieces.first() {
            Some(c) => c,
            None => {
                return full_command;
            },
        };
        let command = String::from(*command_temp);

        pieces.remove(0);
        let arg = match pieces.len() {
            0 => String::from(""),
            _ => format!(" {}", pieces.join(" ")),
        };

        if conf.aliases.contains_key(&command) && command.chars().count() > 0 {
            let ret_string = format!("{}{}", conf.aliases[&command], arg);
            return ret_string;
        } else {
            return full_command;
        }
            
    }   

    fn run_cmd(mycommand: &mut String, 
               current_number: &mut i32,
               conf: &mut SqishConf,
               stdout: &mut RawTerminal<Stdout>) {
        //NORMAL CMD
        if (mycommand != &String::from("")) && (mycommand != &String::from("exit")) {
            //Run the command
            let comm = replace_signs(&mycommand);
            RawTerminal::suspend_raw_mode(&stdout);
            let res = handle_input(&comm);
            //Display the results
            match res {
                Ok(_) => (),
                Err(e) => {
                    println!("\r\n{}Got error {}, RC : {:?}{}",
                             color::Fg(color::Red),
                             e.outp,
                             e.rc,
                             color::Fg(color::Reset));
                },
            };
            //Prepare for a new command
            RawTerminal::activate_raw_mode(&stdout);
            mycommand.clear();
            conf.update_prompt(get_curr_history_number());
            write!(stdout, "{}", conf.promptline).unwrap();
            stdout.flush();
            *current_number += 1;
        //EXITTING
        } else if mycommand == &String::from("exit") {
            write!(stdout, "\r\nSayonara\r\n");
            RawTerminal::suspend_raw_mode(&stdout);
            std::process::exit(0);
        //EMPTY LINE
        } else {
            conf.update_prompt(get_curr_history_number());
            write!(stdout, "\r\n{}", conf.promptline).unwrap();
        }
        stdout.flush();

    }

    fn write_letter(mycommand: &mut String, current_pos: &mut usize, max_pos: &mut usize, c: char) {
        let mut stdout = stdout().into_raw_mode().unwrap();
        if *current_pos == *max_pos {
            mycommand.push(c);
            *current_pos += 1;
            *max_pos += 1;
            write!(stdout, "{}", c).unwrap();
        } else if *current_pos < *max_pos {
            //Inserting a char inside the command...
            write!(stdout, "{}", cursor::Save);
            mycommand.insert(*current_pos, c);
            let command_piece = &mycommand.as_str()[*current_pos..];
            write!(stdout, "{}", command_piece);
            stdout.flush();
            write!(stdout, "{}", cursor::Restore);
            write!(stdout, "{}", cursor::Right(1));
            *max_pos += 1;
            *current_pos += 1;
        }
    }

    fn append_prev_arg(mycommand: &mut String, current_pos: &mut usize, max_pos: &mut usize) {
        let prev_arg = get_prev_arg();
        match prev_arg {
            Some(a) => {
                for letter in a.chars() {
                    write_letter(mycommand, current_pos, max_pos, letter);
                }
            },
            None => (),
        };
    }


    fn find_next_space(current_pos: &usize, max_pos: &usize, command: &String) -> usize {
        let markers = vec![' ', '!', ',', ';', ':'];
        let mut steps: usize = 0;
        if *current_pos == *max_pos { return steps; }
        for letter in command.chars().skip(*current_pos+1) {
            if markers.contains(&letter) {
                if *current_pos + steps < *max_pos { steps += 1; }
                return steps;
            } else {
                steps += 1;
            }
        }
        let max: usize = (0 + max_pos) - current_pos;
        return max;
    }
        

    fn handle_conf() -> SqishConf {
        let conf = match SqishConf::from_sqshrc() {
            Ok(c) => c,
            Err(_) => {
                let conf = SqishConf {
                    promptline: String::from("!$HISTNUMBER$COLORGREEN_[$USER_@$HOSTNAME_]$COLORRESET_ "),
                    promptline_base: String::from("!$HISTNUMBER_$COLORGREEN_[$USER_@$HOSTNAME_]$COLORRESET_ "),
                    aliases: HashMap::new(),
                    hotkeys: HashMap::new(),
                    init: String::new(),
                    env: HashMap::new(),
                };
                conf
            },
        };
        return conf;
    }

    fn set_ctrlc() {
        ctrlc::set_handler(|| {
            ();
        }).unwrap();
    }

    fn tab(elems: &mut TermElements) {
        //Do NOT search on an empty command
        if *&elems.command.len() == 0 {
            return;
        }
        //Search
        *&mut elems.command = replace_signs(&elems.command);
        let (res, list) = Search::build(&elems.command).unwrap().autocomplete();
        if list.len() > 1 { write!(elems.stdout, "\r\n{}", list); }
        elems.command.clear();
        elems.command.push_str(&res);
        elems.max_pos = res.len();
        elems.cur_pos = elems.max_pos;
        write!(elems.stdout, "\r\n{}{}", elems.conf.promptline, res);
        elems.stdout.flush();
    }

    fn enter(elems: &mut TermElements) {
        elems.current_number = get_curr_history_number();
        elems.cur_pos = 0;
        elems.max_pos = 0;
        *&mut elems.command = transform_alias(&elems.conf, elems.command.clone());
        run_cmd(&mut elems.command,
                &mut elems.current_number, 
                &mut elems.conf,
                &mut elems.stdout);
    }

    fn cmd_clear(elems: &mut TermElements) {
        elems.cur_pos = 0;
        elems.max_pos = 0;
        elems.current_number = get_curr_history_number();
        elems.command.clear();
        write!(elems.stdout, "\r\n--CANCEL--\r\n");
        write!(elems.stdout, "{}", elems.conf.promptline);
    }

    fn undo_line(elems: &mut TermElements) {
        clear_line(&elems.max_pos, &elems.cur_pos);
        elems.max_pos = 0;
        elems.cur_pos = 0;
        elems.command.clear();
    }

    fn jmp_start(elems: &mut TermElements) {
        if elems.cur_pos != 0 {
            write!(elems.stdout, "{}", cursor::Left(elems.cur_pos as u16));
            elems.cur_pos = 0;
        }
    }

    fn jmp_end(elems: &mut TermElements) {
        if elems.cur_pos != elems.max_pos {
            let chars_left = elems.max_pos - elems.cur_pos;
            write!(elems.stdout, "{}", cursor::Right(chars_left as u16));
            elems.cur_pos = elems.max_pos;
        }
    }

    fn press_del(elems: &mut TermElements) {
        if elems.command.chars().count() == 1
            && elems.cur_pos == 0 {
                clear_line(&elems.max_pos, &elems.cur_pos);
                elems.command.clear();
                elems.max_pos = 0;
                elems.cur_pos = 0;
        } else if elems.cur_pos < elems.max_pos {
            elems.command.remove(elems.cur_pos);
            if elems.cur_pos > 0 {
                elems.cur_pos -= 1;
                write!(elems.stdout, "{}", cursor::Left(1));
            }
            elems.max_pos -= 1;
            write!(elems.stdout, "{}", cursor::Save);
            clear_line(&elems.max_pos, &elems.cur_pos);
            write!(elems.stdout, "{}", elems.command);
            write!(elems.stdout, "{}", cursor::Restore);
            elems.stdout.flush();
        }
    }

    fn backspace(elems: &mut TermElements) {
        if elems.cur_pos == 0 {
            return;
        }
        if elems.command.chars().count() == 1 {
            clear_line(&elems.max_pos, &elems.cur_pos);
            elems.command.clear();
            elems.cur_pos = 0;
            elems.max_pos = 0;
        } else {
            elems.cur_pos -= 1;
            elems.max_pos -= 1;
            elems.command.remove(elems.cur_pos);
            write!(elems.stdout, "{}", cursor::Left(1));
            write!(elems.stdout, "{}", cursor::Save);
            clear_line(&elems.max_pos, &elems.cur_pos);
            write!(elems.stdout, "{}", elems.command);
            write!(elems.stdout, "{}", cursor::Restore);
            elems.stdout.flush();
       }
    }

    fn uparrow(elems: &mut TermElements) {
        if elems.current_number > 2 {
            elems.command = delete_current_cmd(elems.command.clone());
            elems.current_number -= 1;
            let lastcmd = match get_hist_from_number(&elems.current_number) {
                Some(c) => c,
                None => return
            };
            elems.command = lastcmd.trim().to_string();
            (elems.max_pos, elems.cur_pos) = get_cmd_curs_pos(&elems.command);
            write!(elems.stdout, "{}", elems.command);
        }
    }

    fn downarrow(elems: &mut TermElements) {
        if elems.current_number < get_curr_history_number() {
            elems.command = delete_current_cmd(elems.command.clone());
            elems.current_number += 1;
            let lastcmd = match get_hist_from_number(&elems.current_number) {
                Some(c) => c,
                None => return
            };
            elems.command = lastcmd.trim().to_string();
            (elems.max_pos, elems.cur_pos) = get_cmd_curs_pos(&elems.command);
            write!(elems.stdout, "{}", elems.command);
        }
    }

    fn leftarrow(elems: &mut TermElements) {
        if elems.cur_pos > 0 {
            elems.cur_pos -= 1;
            write!(elems.stdout, "{}", cursor::Left(1));
        }
    }

    fn rightarrow(elems: &mut TermElements) {
        if elems.cur_pos < elems.max_pos {
            print!("{}", cursor::Right(1));
            elems.cur_pos += 1;
        }
    }

    fn screen_clear(elems: &mut TermElements) {
        write!(elems.stdout, "{}{}", termion::clear::All, termion::cursor::Goto(1,1));
        write!(elems.stdout, "{}{}", elems.conf.promptline, elems.command);
    }

    fn jmp_nxt_word(elems: &mut TermElements) {
        let steps = find_next_space(&elems.cur_pos, &elems.max_pos, &elems.command);
        //println!("steps: {:?} cur {:?} max {:?}", steps, elems.cur_pos, elems.max_pos);
        if steps > 0 {
            print!("{}", cursor::Right(steps as u16));
            elems.cur_pos += steps;
        }
    }

    fn jmp_prev_word(elems: &mut TermElements) {
        let command_rev = elems.command.chars().rev().collect::<String>();
        let curr_rev = elems.max_pos - elems.cur_pos;
        let steps = find_next_space(&curr_rev, &elems.max_pos, &command_rev);
        if steps > 0 {
            print!("{}", cursor::Left(steps as u16));
            elems.cur_pos -= steps;
        }
    }

    fn press_alt(elems: &mut TermElements, x: char) {
        match x {
            'a'..='z' => {
                let x = String::from(x);
                if elems.conf.hotkeys.contains_key(&x) {
                    let hotcmd = &elems.conf.hotkeys[&x].clone();
                    for c in hotcmd.chars() {
                        if c != '\n' {
                            write_letter(&mut elems.command,
                                        &mut elems.cur_pos,
                                        &mut elems.max_pos,
                                        c);
                        } else {
                            elems.cur_pos = 0;
                            elems.max_pos = 0;
                            elems.command = transform_alias(&elems.conf, elems.command.clone());
                            run_cmd(&mut elems.command,
                                &mut elems.current_number,
                                &mut elems.conf,
                                &mut elems.stdout);
                        }

                    }
                }
            },
            '.' => {
                append_prev_arg(&mut elems.command, &mut elems.cur_pos, &mut elems.max_pos);
            },
            '!' => {
                let mut cmd = format!("\r\n-- HOTKEYS --> \r\n{:?}\r\n", elems.conf.hotkeys);
                cmd = format!("{}\r\n-- ALIASES -->\r\n{:?}\r\n", cmd, elems.conf.aliases);
                cmd = cmd.replace(",", "\r\n");
                cmd = cmd.replace('"', "");
                write!(elems.stdout, "{}", cmd);
            },
            _ => (),
        };
    }

    //I smell horrible code in here
    fn set_envvars(vars: &HashMap<String, String>) {
        let vars = vars.clone();
        let re = Regex::new(r"\$[A-Za-z_]+").unwrap();

        for (key, value) in vars {
            let mut after = value.clone();
            let mut nbr_of_vars = *&value.matches('$').count();
            while nbr_of_vars >= 1 {
                let temp = after.clone();
                let caps = re.captures(&temp).unwrap();
                for cap in caps.iter() {
                    match cap {
                        Some(c) => {
                            let myvar = String::from(c.as_str()).replace('$', "");
                            match env::var(myvar) {
                                Ok(r) => {
                                    *&mut after = after.replace(c.as_str(), &r);
                                },
                                Err(_) => continue,
                            };
                        },
                        None => continue,
                    };
                }
                nbr_of_vars -= 1;
            }
            env::set_var(&key, &after);
        }
    }


    //THE ENTRYPOINT FUNCTION
    pub fn run_raw() {

        let mut elems = TermElements {
            command: String::new(),
            cur_pos: 0,
            max_pos: 0,
            stdout: stdout().into_raw_mode().unwrap(),
            current_number: get_curr_history_number(),
            conf: handle_conf(),
        };

        let stdin = stdin();

        //Handle Ctrl+C
        set_ctrlc();

        &elems.conf.update_prompt(get_curr_history_number());

        //Initializing
        //write!(elems.stdout, "\r\n ---Sqish initializing--- \r\n{}", elems.conf.promptline);
        elems.stdout.flush();
        set_envvars(&elems.conf.env);

        //Read hotkeys set as environment variables
        elems.conf.hotkeys = read_hotkeys(&elems.conf.hotkeys);

        if elems.conf.init != String::from("") {
            run_cmd(&mut String::from(&elems.conf.init), &mut elems.current_number, &mut elems.conf, &mut elems.stdout);
        } else {
            run_cmd(&mut String::from(""), &mut elems.current_number, &mut elems.conf, &mut elems.stdout);
        }

        for c in stdin.keys() {
            let k = c.unwrap();
            match k {
               Key::Char('\t') => {
                   tab(&mut elems);
                }
                Key::Char('\n') => {
                    enter(&mut elems);
                    elems.conf.hotkeys = read_hotkeys(&elems.conf.hotkeys);
                },

                Key::Ctrl('q'|'d') => {
                    writeln!(elems.stdout, "\r\nSayonara\r\n");
                    break;
                },

                Key::Ctrl('c') => {
                    cmd_clear(&mut elems);
                },

                Key::Ctrl('u') => {
                    undo_line(&mut elems);
                },

                Key::Ctrl('a') => {
                    jmp_start(&mut elems);
                },

                Key::Ctrl('e') => {
                    jmp_end(&mut elems);
                },

                Key::Char(c) => {
                    write_letter(&mut elems.command, &mut elems.cur_pos, &mut elems.max_pos, c);
                }

                Key::Delete => {
                    press_del(&mut elems);
                },

                Key::Backspace => {
                    backspace(&mut elems);
                },

                Key::Up => {
                    uparrow(&mut elems);
                },

                Key::Down => {
                    downarrow(&mut elems);
                },

                Key::Left => {
                    leftarrow(&mut elems);
                },

                Key::Right => {
                    rightarrow(&mut elems);
                },

                Key::Ctrl('l') => {
                    screen_clear(&mut elems);
                },

                Key::Ctrl('p') => {
                    jmp_nxt_word(&mut elems);
                },

                Key::Ctrl('o') => {
                    jmp_prev_word(&mut elems);
                },
                
                Key::Alt(x) => {
                    press_alt(&mut elems, x);
                },

                _ => (),
            };

            elems.stdout.flush().unwrap();
        }

    }
}
